//---------------------------------------------------------------------------

#pragma hdrstop

#include <condefs.h>

//---------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>

#include "FTIndex.h"

#define NUM_PATH_PTS	6
ttPath	vrJur[NUM_PATH_PTS] =  {{113.0,10.0},{109.0,13.0},{101.0,14.0},
													{96.0,17.0},{5.0,11.5},{0.0,10.5}};
ttPath	vrCret[NUM_PATH_PTS] = {{113.0,10.0},{109.0,16.0},{101.0,16.0},
													{96.0,21.0},{5.0,12.5},{0.0,11.5}};
ttPath	vrMean[NUM_PATH_PTS] = {{113.0,10.0},{109.0,14.5},{101.0,15.0},
													{96.0,19.0},{5.0,12.0},{0.0,11.0}};
ttPath	fcPath[NUM_PATH_PTS] = {{27.8,20.0},{25.8,20.0},{20.0,20.0},
                                       {15.0,20.0},{10.0,20.0},{0.0,20.0}};


/*  RcToDensity
		Does the conversion from length to density for a c-axis-projected length,
    based on Ketcham et al., 1999 model.
		Also used to estimate	bias for population summing.
 */
double RcToDensity(double cParLen)
{
/*  return(RmToDensity(-1.499*cParLen*cParLen + 4.15*cParLen - 1.656)); */
	if (cParLen > 0.757) return(1.600*cParLen-0.599);
	if (cParLen >= 0.55) return(9.205*cParLen*cParLen-9.157*cParLen+2.269);
	return(0.0);
}

/*	RmToDensity
    Estimates density for a population given its reduced mean length.
    This is used both for direct denisty estimation and for estimating the
    relative probability of this population being observed.
    The latter usage should probably be revised by recasting directly as track loss.
 */
double	RmToDensity(double redLen)
{
  if (redLen >= 0.66642) return(redLen*1.057398 - 0.057398);
  if (redLen >= 0.64402) return(0.54105+0.10622*(redLen-0.64402)/0.02239);
  if (redLen >= 0.40356) return(0.12628+0.414768*(redLen-0.40356)/0.240467);
  return(redLen*0.31292);
}


/* TempFading
	Returns the temperature at which fission tracks fade (mean length falls
   to FADING_LENGTH or FADING_LENGTH_C) after "timeMa" million years at
   constant temperature, given parameters p.
 */
double TempFading(double timeMa, double *p,double rmr0,double k,int fitType,int lenType)
{
	double minLen,calc;
  double loT,hiT,midT,len;

	if (fitType != F_LAS_GAL) {
    minLen = (lenType == D_MEAN ? RM_TOTAL_ANNEAL : RC_TOTAL_ANNEAL);
		calc = pow(minLen,1.0/k)*(1.0-rmr0)+rmr0;
  }
	switch (fitType) {
   	case F_FANNING_6:
      calc = (pow((1.0-pow(calc,p[6]))/p[6],p[5])-1.0)/p[5] - p[1];
      calc = p[2]*(log(SECS_PER_MA*timeMa)-p[3])/calc + p[4];
      calc = 1.0/calc - 273.15;
      return(calc);
    case F_FANNING_CURVE:
      calc = (pow((1.0-pow(calc,p[6]))/p[6],p[5])-1.0)/p[5] - p[1];
      calc = p[2]*(log(SECS_PER_MA*timeMa)-p[3])/calc + p[4];
      calc = 1.0/exp(calc) - 273.15;
      return(calc);
    case F_LAS_GAL:
      calc = log(1.0-pow(7.0/p[1],1.0/p[6])) - p[2];
      calc = p[3]*(log(SECS_PER_MA*timeMa/SECS_PER_HOUR)-p[4])/calc + p[5];
      calc = 1.0/calc - 273.15;
      return(calc);
    case F_CARLSON:
    /* Because T appears twice, need to iterate to find solution */
      loT = 0.0 + 273.15;
      hiT = 250.0 + 273.15;
      while (hiT-loT > 0.01) {
        midT = (loT + hiT) * 0.5;
        len = p[1]*pow(BOLTZMANN*midT/PLANCK,p[2])*exp(-p[2]*p[3]/(GAS_CONST*KCAL_PER_J*midT));
        len = p[4] - len*pow(timeMa*SECS_PER_MA,p[2]);
        if (len < 12.0) len = len/(1.0+0.1*(12.0-len));
        len = len/p[4];
        if (len < calc) hiT = midT;
        else loT = midT;
      }
      return(midT-273.15);
    default:
      return(0);
  }
}

/* ClosureAndAnnealingTemp
	Returns the closure temperature and temperature of total annealing for an
   annealing model.
 */
void	ClosureAndAnnealingTemp(double coolRate,double *p,double rmr0,double k,
										double *tc,double *ta,int fitType,int lenType)
{
	const double startTemp = 20.0;    // Initial temperature is 20 deg C
	double currTemp = startTemp + 273.15;
  double currLength = 0.9999;
  double tempLength;
  double timeElapsed = 0.0;
  double age = 0.0;
  double equivTime,timeInt,meanTemp;
  const double tempStep = 0.01;      // Temp step is 0.01 deg C - small enough!
  double minLen = (lenType == D_MEAN ? RM_TOTAL_ANNEAL : RC_TOTAL_ANNEAL);
  double timeStep = tempStep/coolRate;
  double timeStepSec = timeStep * SECS_PER_MA;

  if (fitType == F_LAS_GAL) {
   	currLength = p[1]-0.01;
    minLen = 7.0;
  }
  if (fitType == F_CARLSON) {
    currLength = p[4]-0.01;
    minLen *= p[4];
    minLen = (minLen + 0.1*12.0*minLen)/(0.1*minLen + 1.0);
  }

  while (currLength > minLen) {
   	meanTemp = currTemp + tempStep/2.0;
   	currTemp += tempStep;
    timeElapsed += timeStep;
    if (fitType == F_CARLSON) {
      equivTime = (p[4]-currLength)/(p[1]*pow(BOLTZMANN*meanTemp/PLANCK,p[2]));
      equivTime /= exp(-p[3]*p[2]/(GAS_CONST*KCAL_PER_J*meanTemp));
      equivTime = pow(equivTime,1.0/p[2]);
    } else {
      equivTime = pow(currLength,1.0/k)*(1-rmr0) + rmr0;
      if (fitType == F_LAS_GAL) {
        equivTime = log(1.0-pow(equivTime/p[1],1.0/p[6]))-p[2];
        equivTime = equivTime*(1.0/meanTemp-p[5])/p[3] + p[4];
        equivTime = exp(equivTime)*SECS_PER_HOUR;
      } else {
   			equivTime = pow((1.0-pow(equivTime,p[6]))/p[6],p[5]);
        equivTime = ((equivTime - 1.0)/p[5] - p[1])/p[2];
        if (fitType == F_FANNING_6)
          equivTime = exp(equivTime*(1.0/meanTemp-p[4])+p[3]);
        else if (fitType == F_FANNING_CURVE)
          equivTime = exp(equivTime*(log(1.0/meanTemp)-p[4])+p[3]);
      }
    }
    timeInt = equivTime + timeStepSec;
    if (fitType == F_LAS_GAL) {
      currLength = 1.0-exp(p[2]+p[3]*(log(timeInt/SECS_PER_HOUR)-p[4])/(1.0/meanTemp-p[5]));
      currLength = p[1]*pow(currLength,p[6]);
      currLength = (currLength < 0) ? 0.0 : currLength;
    } else if (fitType == F_CARLSON) {
      currLength = p[1]*pow(BOLTZMANN*meanTemp/PLANCK,p[2]);
      currLength *= exp(-p[2]*p[3]/(GAS_CONST*KCAL_PER_J*meanTemp));
      currLength = p[4] - currLength*pow(timeInt,p[2]);
    } else {
      if (fitType == F_FANNING_6)
				currLength = 1.0+p[5]*(p[1]+p[2]*(log(timeInt)-p[3])/(1.0/meanTemp-p[4]));
			else
				currLength = 1.0+p[5]*(p[1]+p[2]*(log(timeInt)-p[3])/(log(1.0/meanTemp)-p[4]));
			currLength = 1.0-p[6]*pow(currLength,1.0/p[5]);
			currLength = (currLength < 0) ? 0.0 : pow(currLength, 1.0/p[6]);
    }
    currLength = pow((currLength - rmr0)/(1 - rmr0),k);

    if (fitType == F_LAS_GAL)
      age += timeStep*RmToDensity(currLength/16.33);
    else if (fitType == F_CARLSON) {
      tempLength = currLength;
    // Convert to segmented length if necessary
      if (currLength < 12.0) tempLength = currLength/(1.0+0.1*(12.0-currLength));
      age += timeStep*RmToDensity(tempLength/p[4]);
    }
    else {
      if (lenType == D_MEAN) age += timeStep*RmToDensity(currLength);
   		else age += timeStep*RcToDensity(currLength);
    }
  }
  *ta = currTemp - 273.15;
  age /= 0.893;  /* Standard length reduction */
  *tc = 20.0+age*coolRate;
}

double VrolijkLength(double *p,double rmr0,double k,ttPath *path,int fitType,int lenType)
{
	double currTemp,tempStep,currSlope;
  int currPt,currPop;
  double timeStep = 1.0;			/* Calculate a population every 1 m.y. */
  double currLength = 0.9999;
  double segLength;
  double lengths[114];
  double currTime = 0.0;
  double equivTime,timeInt,meanTemp;
  double timeStepSec = timeStep * SECS_PER_MA;
  double lenSum,weightSum,weight,obsBias;

	currPt = NUM_PATH_PTS-1;
  currPop = 0;
  if (fitType == F_LAS_GAL) currLength = p[1]-0.01;
  if (fitType == F_CARLSON) currLength = p[4]-0.01;

  currSlope = (path[currPt].temp-path[currPt-1].temp)/(path[currPt].time-path[currPt-1].time);
  tempStep = currSlope * timeStep;
  currTemp = path[currPt].temp + 273.15;
  while (currPt) {     /* THIS LOOP ASSUMES WE DON'T FULLY ANNEAL ANYTHING! */
   	meanTemp = currTemp + tempStep/2.0;
   	currTemp += tempStep;
    currTime += timeStep;
    equivTime = pow(currLength,1.0/k)*(1-rmr0) + rmr0;

    if (fitType == F_LAS_GAL) {
      equivTime = log(1.0-pow(equivTime/p[1],1.0/p[6]))-p[2];
      equivTime = equivTime*(1.0/meanTemp-p[5])/p[3] + p[4];
      equivTime = exp(equivTime)*SECS_PER_HOUR;
    } else if (fitType == F_CARLSON) {
      equivTime = (p[4]-currLength)/(p[1]*pow(BOLTZMANN*meanTemp/PLANCK,p[2]));
      equivTime /= exp(-p[3]*p[2]/(GAS_CONST*KCAL_PER_J*meanTemp));
      equivTime = pow(equivTime,1.0/p[2]);
    } else {
   		equivTime = pow((1.0-pow(equivTime,p[6]))/p[6],p[5]);
      equivTime = ((equivTime - 1.0)/p[5] - p[1])/p[2];
      if (fitType == F_FANNING_6)
      	equivTime = exp(equivTime*(1.0/meanTemp-p[4])+p[3]);
      else if (fitType == F_FANNING_CURVE)
        equivTime = exp(equivTime*(log(1.0/meanTemp)-p[4])+p[3]);
    }
    timeInt = equivTime + timeStepSec;
    if (fitType == F_LAS_GAL) {
      currLength = 1.0-exp(p[2]+p[3]*(log(timeInt/SECS_PER_HOUR)-p[4])/(1.0/meanTemp-p[5]));
      currLength = p[1]*pow(currLength,p[6]);
      currLength = (currLength < 0) ? 0.0 : currLength;
    } else if (fitType == F_CARLSON) {
      currLength = p[1]*pow(BOLTZMANN*meanTemp/PLANCK,p[2]);
      currLength *= exp(-p[2]*p[3]/(GAS_CONST*KCAL_PER_J*meanTemp));
      currLength = p[4] - currLength*pow(timeInt,p[2]);
    } else {
      if (fitType == F_FANNING_6)
				currLength = 1.0+p[5]*(p[1]+p[2]*(log(timeInt)-p[3])/(1.0/meanTemp-p[4]));
			else
				currLength = 1.0+p[5]*(p[1]+p[2]*(log(timeInt)-p[3])/(log(1.0/meanTemp)-p[4]));
			currLength = 1.0-p[6]*pow(currLength,1.0/p[5]);
			currLength = (currLength < 0) ? 0.0 : pow(currLength, 1.0/p[6]);
    }
    currLength = pow((currLength - rmr0)/(1 - rmr0),k);
    if (fitType == F_LAS_GAL) lengths[currPop++] = currLength/16.33;
    else if (fitType == F_CARLSON) {
      segLength = currLength;
      if (currLength < 12.0) segLength = segLength/(1.0+0.1*(12.0-segLength));
      lengths[currPop++] = segLength/p[4];
    } else lengths[currPop++] = currLength;
    if (currTime+timeStep/2.0 > path[currPt-1].time) { /* Hack to avoid roundoff */
      currPt--;
      if (currPt) {
   			currSlope = (path[currPt].temp-path[currPt-1].temp)/(path[currPt].time-path[currPt-1].time);
   			tempStep = currSlope * timeStep;
   			currTemp = path[currPt].temp + 273.15;
      }
    }
  }
/* Length calculation done */

  for (currPop=0,lenSum=0,weightSum=0;currPop<113;currPop++) {
   	/* weighting for U loss */
    weight = (exp(U238MA*(currPop+1))-exp(U238MA*(currPop)))/U238MA;
    /* weighting for observational bias */
    obsBias = (lenType == D_MEAN) ? RmToDensity(lengths[currPop]) : RcToDensity(lengths[currPop]);
    weight *= obsBias;
    lenSum += lengths[currPop] * weight;
    weightSum += weight;
  }
  return(lenSum/weightSum);
}

/* CalcBenchmarks
  Organizes calculation of benchmarks for a given annealing model.

  Inputs:
    FILE *repFile: Text file for writing results
    int  fitType:
 */
void CalcBenchmarks(FILE *repFile,int fitType,int lenType,double *results,
							char *dataFileName,double rmr0,double k)
{
	double tF10,tF30,tF100,tC1,tC10,tC100,tA1,tA10,tA100,lVrolC,lVrolJ,lVrolMean,lFC;

  printf("Calculating benchmark functions\n");

  tF10 = TempFading(10.0,results,rmr0,k,fitType,lenType);
  tF30 = TempFading(30.0,results,rmr0,k,fitType,lenType);
  tF100 = TempFading(100.0,results,rmr0,k,fitType,lenType);
  ClosureAndAnnealingTemp(1.0,results,rmr0,k,&tC1,&tA1,fitType,lenType);
  ClosureAndAnnealingTemp(10.0,results,rmr0,k,&tC10,&tA10,fitType,lenType);
  ClosureAndAnnealingTemp(100.0,results,rmr0,k,&tC100,&tA100,fitType,lenType);
	lVrolC = VrolijkLength(results,rmr0,k,vrCret,fitType,lenType);
	lVrolJ = VrolijkLength(results,rmr0,k,vrJur,fitType,lenType);
	lVrolMean = VrolijkLength(results,rmr0,k,vrMean,fitType,lenType);
  lFC = VrolijkLength(results,rmr0,k,fcPath,fitType,lenType);

  fprintf(repFile,"%s\t%.4lf\t%.4lf\t",dataFileName,rmr0,k);
  fprintf(repFile,"%.2lf\t%.2lf\t%.2lf",tF100,tF30,tF10);
  fprintf(repFile,"\t%.2lf\t%.2lf\t%.2lf",tC1,tC10,tC100);
  fprintf(repFile,"\t%.2lf\t%.2lf\t%.2lf",tA1,tA10,tA100);
  fprintf(repFile,"\t%.4lf\t%.4lf\t%.4lf\t%.4lf\n",lVrolC,lVrolJ,lVrolMean,lFC);
  printf("%s\t%.2lf\t%.2lf\t%.2lf\t%.2lf\t%.2lf\t%.2lf\t%.2lf\t%.2lf\t%.2lf\t%.4lf\t%.4lf\t%.4lf\n",
   	dataFileName,tF100,tF30,tF10,tC1,tC10,tC100,tA1,tA10,tA100,lVrolC,lVrolJ,lVrolMean);
}


void RunBenchmarks(void)
{
  char  inFileName[255],outFileName[255], modelName[255];
  double  c[7],rmr0,k;
  int   modelType,lenType;
  FILE *fpIn, *fpOut;

  printf("Enter the name of the file with model parameters\n");
  scanf("%s",inFileName);
	if ((fpIn = fopen(inFileName,"r")) == NULL) {
		perror(inFileName);
		return;
	}

  printf("Enter the name for output file\n");
  scanf("%s",outFileName);
	if ((fpOut = fopen(outFileName,"w")) == NULL) {
		perror(outFileName);
		return;
	}
  fprintf(fpOut,"Model\trmr0\tkappa\tTF,100\tTF,30\tTF,10\tTC,1\tTC,10\tTC,100\tTA,1\tTA,10\tTA,100\tLvrolC\tLvrolJ\tlVrolMean\tlFC\n");

  /* Skip header line in input file */
  while (getc(fpIn) != '\n') ;

  while (fscanf(fpIn,"%s %d %d",modelName,&modelType,&lenType) != EOF) {
    fscanf(fpIn,"%lf %lf %lf %lf %lf %lf",&rmr0, &k, &c[1], &c[2], &c[3], &c[4]);
    if (modelType != F_CARLSON) fscanf(fpIn,"%lf %lf",&c[5], &c[6]);
    CalcBenchmarks(fpOut,modelType,lenType,c,modelName,rmr0,k);
  }
	fclose(fpOut);
  fclose(fpIn);
}

#pragma argsused
int main(int argc, char* argv[])
{
  RunBenchmarks();

  return 0;
}
//---------------------------------------------------------------------------
 